//! 
//! This is a pared down version of the main load-wasmer program without
//! all of the bells and whistles. Good for tracking down bugs and minimizing examples.

use std::fs;
use structopt::StructOpt;
use wasmer::{imports, Instance, Module, NativeFunc, Store, Function};
use wasmer_compiler_cranelift::{Cranelift, CraneliftOptLevel};
use wasmer_compiler_llvm::{LLVM, LLVMOptLevel};
use wasmer_engine_native::Native;


/// Run a wasm program using Wasmer
#[derive(StructOpt)]
struct Cli {
    /// The path to the wasm program (.wasm).
    #[structopt(required(true))]
    path: std::path::PathBuf,
}


fn print_val_i32(val: i32) -> i32 {
    println!("[i32]   {}", val);
    return val;
}


fn main() -> anyhow::Result<()> {
    let args = Cli::from_args();
  
    // Insantiate the compiler and engine into a store
    let mut cranelift_compiler = Cranelift::new();
    cranelift_compiler.opt_level(CraneliftOptLevel::None);
    let cranelift_store = Store::new(&Native::new(cranelift_compiler).engine());

    let mut llvm_compiler = LLVM::new();
    llvm_compiler.opt_level(LLVMOptLevel::None);
    let llvm_store = Store::new(&Native::new(llvm_compiler).engine());


    let cranelift_wasm_bytes = fs::read(args.path.clone())?;
    let llvm_wasm_bytes = fs::read(args.path.clone())?;
  
    let cranelift_module = Module::new(&cranelift_store, cranelift_wasm_bytes)?;
    let llvm_module = Module::new(&llvm_store, llvm_wasm_bytes)?;

    let cranelift_print_val_i32_func = Function::new_native(&cranelift_store, print_val_i32);
    let llvm_print_val_i32_func = Function::new_native(&llvm_store, print_val_i32);

    // Create an import object. Has no effect if wasm doesn't import it
    let cranelift_import_object = imports! {
        "env" => {
           "print_i32" => cranelift_print_val_i32_func, 
        },
    };

    let llvm_import_object = imports! {
        "env" => {
           "print_i32" => llvm_print_val_i32_func, 
        },
    };
    // Let's instantiate the Wasm module.
    let cranelift_instance = Instance::new(&cranelift_module, &cranelift_import_object)?;
    let llvm_instance = Instance::new(&llvm_module, &llvm_import_object)?;

    // Get handles for the exports
    let cranelift_main_func: NativeFunc<(), i32> = cranelift_instance.exports.get_native_function("_main")?;
    let llvm_main_func: NativeFunc<(), i32> = llvm_instance.exports.get_native_function("_main")?;

    // Call main and display the result
    let cranelift_result = cranelift_main_func.call()?;
    let llvm_result = llvm_main_func.call()?;
    println!("Result from cranelift: {}", cranelift_result);
    println!("Result from llvm: {}", llvm_result);


    Ok(())
}
